"""
A set of variables used for testing the relayr API.
Attention: You must provide your own eaningful values 
#before you can run large parts of the test suite!
"""

testset0 = {
    'userEmail': 'god@in.heaven',
    'relayrAPI': 'https://api.relayr.io'
}



    # ..............................................................................
    # System
    # ..............................................................................

    def post_client_log(self, log_messages):
        """
        Start logging (experimental).

        Sample data input::

            [{
                "timestamp"  : "Time stamp in ISO 8601",
                "message" : "The message log by the developer",
                "connection" : {
                    "internet" : True,
                    "netScope" : "WAN",
                    "netType"  : "LTE"
                }
             },
             {
                ...
             }
            ]
        """
        log_message = json.dumps(log_messages)
        # https://api.relayr.io/client/log
        url = '{0}/client/log'.format(self.host)
        _, data = self.perform_request('POST', url, data=log_messages, headers=self.headers)
        return data



# -*- coding: utf-8 -*-

"""
A set of configuration variables used for various purposes.

This module provides defaults for configuration variables and a way
of overwriting these from the shell environment or a script using
the relayr client.
"""

import os
import platform

from .version import __version__


# defaults
relayrAPI = 'https://api.relayr.io'
clientName = 'io.relayr.sdk.python'
dataConnectionHubName = 'PubNub'
userAgentString = '{client_name}/{client_version} '
userAgentString += '({platform}; {arch}; {python_implementation}-{python_version})'
DEBUG = False
LOG = False
LOG_DIR = os.getcwd()

# overwrite with environment variables if given
relayrAPI = os.environ.get('RELAYR_API', relayrAPI)
clientName = os.environ.get('RELAYR_PYTHON_CLIENT_NAME', clientName)
dataConnectionHubName = os.environ.get('RELAYR_DATAHUB', dataConnectionHubName)
userAgentString = os.environ.get('RELAYR_DATAHUB', userAgentString)
DEBUG = True if os.environ.get('RELAYR_DEBUG', 'False') == 'True' else False
LOG = True if os.environ.get('RELAYR_LOG', 'False') == 'True' else False
LOG_DIR = os.environ.get('RELAYR_LOG_DIR', LOG_DIR)

# derived variable, HTTP user-agent string
userAgent = userAgentString.format(
	client_name=clientName,
	client_version=__version__,
	platform=platform.system() + '-' + platform.release(),
	arch=platform.machine() + '-' + platform.architecture()[0],
	python_implementation=platform.python_implementation(),
	python_version=platform.python_version(),
)

del os, platform



# -*- coding: utf-8 -*-

"""
This module contains tests of the Relayr restful HTTP API client.

Attention: The values in these fixture files need to be set properly before
running the tests!
"""

import pytest


class TestPubNub(object):
    "Test accessing device data via a PubNub connection."

    def receive(self, message, channel):
        "Callback method for data connection."
        print(message) # suppressed by py.test
        self.received_data.append(message)

    def test_read_pubnub_device_10s(self, fix_registered):
        "Test connect to a device via PubNub and read data for some time."
        import time
        from relayr import Client
        from relayr.resources import Device
        token = fix_registered.testset1['token']
        deviceID = fix_registered.testset1['deviceID']
        self.received_data = []
        c = Client(token=token)
        dev = Device(id=deviceID, client=c).get_info()
        usr = c.get_user()
        app = c.get_app()
        conn = usr.connect_device(app, dev, self.receive)
        conn.start()
        time.sleep(10)
        conn.stop()
        assert len(self.received_data) > 0
        for item in self.received_data:
            assert 'ts' in item



# -*- coding: utf-8 -*-

"""
This module contains exceptions raised when an operation on the 
relayr API or the relayr platform fails due to, 
for example: missing credentials, invalid UIDs, etc.

At the moment two exception classes are provided: 

- ``RelayrApiException``: raised for exceptions caused by API calls
- ``RelayrException``: raised for other exceptions
"""


class RelayrApiException(Exception):
    """
    RelayrApiException
    """

class RelayrException(Exception):
    """
    RelayrException
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Show current Relayr API key figures on the command-line.

This script uses the Python bindings for the Relayr RESTful API as defined
in the ``relayr`` Python package. It should work as-is on Python 2 and 3.

Example:

$ python api_pulse.py 
Relayr API Pulse
............................................................................
Path: demos/api_pulse.py
MD5 hash: 1507cfcb103b8e5df5d301c206138b93
API: https://api.relayr.io
Python Relayr client version: 0.1.1
Python version: CPython 2.7.6
Platform: Darwin-10.8.0-i386-64bit
Date/Time: 2014-10-15T14:13:51.480319
............................................................................
server status ok: True (1.02 s)
317 publishers (0.64 s)
199 applications (0.74 s)
109 public devices (0.87 s)
8 device models (0.42 s)
8 device model meanings (0.47 s)
#public devices by meaning: {
    "acceleration": 17,
    "temperature": 33,
    "noise_level": 18,
    "angular_speed": 17,
    "color": 28,
    "luminosity": 28,
    "proximity": 28,
    "humidity": 33
}
"""

import sys
import json
import time
import datetime
import platform
from hashlib import md5

import termcolor

from relayr import Api
from relayr.compat import urlopen, URLError
from relayr.version import __version__ as relayr_version
from relayr.utils.terminalsize import get_terminal_size


def connected_to_internet():
    "Check if we have access to 'the internet', well some Google server."

    try:
        # Give IP address to avoid potential DNS issues.
        response = urlopen('http://74.125.228.100')
        return True
    except URLError as err:
        return False


def show_versions():
    "Show some versioning info."
    
    # Show header lines.
    api = Api()
    url = api.host
    dt = datetime.datetime.now().isoformat()
    term_width = get_terminal_size()[0]
    print('.' * term_width)
    print('Path: %s' % __file__)
    hash = md5(open(__file__).read().encode('utf8')).hexdigest()
    print('MD5 hash: %s' % hash)
    print('API: %s' % url)
    print('Python Relayr client version: %s' % relayr_version)
    pyimpl = platform.python_implementation()
    pyver = platform.python_version()
    platf = platform.platform()
    print('Python version: %s %s' % (pyimpl, pyver))
    print('Platform: %s' % platf)
    print('Date/Time: %s' % dt)
    print('.' * term_width)
    

def show_check():
    "Show some key numbers about the current state of the API."
    
    # For all this no credentials/token etc. are needed.
    api = Api()

    # Get server status.
    t0 = time.time()
    try:
        res = api.get_server_status()
        ok = res['database'] == 'ok'
    except:
        msg = 'Relayr API not reachable, sorry for that!'
        termcolor.cprint(msg, 'red')
        sys.exit(-1)
    t = time.time() - t0
    print('server status ok: %s (%.2f s)' % (ok, t))

    # Get number of all publishers.
    t0 = time.time()
    publishers = list(api.get_public_publishers())
    t = time.time() - t0
    print('%d publishers (%.2f s)' % (len(publishers), t))

    # Get number of all apps.
    t0 = time.time()
    apps = list(api.get_public_apps())
    t = time.time() - t0
    print('%d applications (%.2f s)' % (len(apps), t))
    
    # Get number of all public devices.
    t0 = time.time()
    devices = list(api.get_public_devices())
    t = time.time() - t0
    print('%d public devices (%.2f s)' % (len(devices), t))

    # Get number of all device models.
    t0 = time.time()
    models = list(api.get_public_device_models())
    t = time.time() - t0
    print('%d device models (%.2f s)' % (len(models), t))

    # Get number of all device model meanings.
    t0 = time.time()
    meanings = list(api.get_public_device_model_meanings())
    t = time.time() - t0
    print('%d device model meanings (%.2f s)' % (len(meanings), t))

    # Get a count of all public devices by meaning.
    counts = {}
    for device in devices:
        readings = device['model']['readings']
        for reading in readings:
            meaning = m = reading['meaning']
            counts[m] = counts[m] + 1 if counts.get(m, 0) else 1
    print('#public devices by meaning: %s' % json.dumps(counts, indent=4))


if __name__ == "__main__":
    print('Relayr API Pulse')
    show_versions()
    if not connected_to_internet():
        msg = 'Sorry, but you do not seem to have internet connectivity!'
        termcolor.cprint(msg, 'red')
        sys.exit(-1)    
    show_check()



"""
This module contains test fixture loaders that py.test will find here when
executing test cases.
"""

import os
import sys

import pytest


## TODO: maybe use importlib.import_module

@pytest.fixture(scope='module')
def fix_anonymous():
    "Return fixture for anonymous API access."
    sys.path.insert(0, os.path.dirname(__file__))
    import fixture_anonymous
    del sys.path[0]
    return fixture_anonymous

@pytest.fixture(scope='module')
def fix_registered():
    "Return fixture for access by a registered user."
    sys.path.insert(0, os.path.dirname(__file__))
    import fixture_registered_dcg
    del sys.path[0]
    return fixture_registered_dcg



"""
Compatibility helpers for Python 2 and 3.
"""

import sys


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3
PY26 = sys.version_info[:2] == (2, 6)

if PY2:
    from urllib import urlopen
    from urllib import urlencode
    from urllib2 import URLError
else:
    from urllib.request import urlopen
    from urllib.parse import urlencode
    from urllib.error import URLError



# -*- coding: utf-8 -*-

"""
This module contains configuration tests for the Relayr restful HTTP API client.

Attention: The configuration mechanism is likely to change.
"""

import os
import json
import socket
from threading import Thread
try:
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
except ImportError:
    from http.server import HTTPServer, BaseHTTPRequestHandler

import pytest

from relayr.compat import PY3


def get_free_port():
    "Return a free port number."

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(('', 0))
    port = sock.getsockname()[1]
    return port


class GetHandler(BaseHTTPRequestHandler):
    "A handler that returns the received HTTP header as serialized JSON body."

    def do_GET(self):
        message = json.dumps(dict(self.headers.items()), indent=4)
        self.send_response(200)
        self.end_headers()
        if PY3:
            self.wfile.write(bytes(message, 'UTF-8'))
        else:
            self.wfile.write(message)


class MyThread(Thread):
    "A tiny HTTP server running in a thread."

    def __init__(self, port):
        Thread.__init__(self)
        self.port = port

    def run(self):
        self.server = HTTPServer(('localhost', self.port), GetHandler)
        self.server.serve_forever()

    def stop(self):
        self.server.shutdown()


class TestConfig(object):
    "Test configuration aspects."

    def test_change_useragent(self):
        "Test changing the HTTP useragent string."

        from relayr import config
        from relayr.api import Api
        port = get_free_port()
        t = MyThread(port)
        t.start()

        previous = config.relayrAPI[:]

        url = 'http://localhost:%d' % port
        config.relayrAPI = url

        if PY3:
            user_agent_key = 'User-Agent'
        else:
            user_agent_key = 'user-agent'

        # default config user-agent
        headers = Api().get_server_status()
        assert headers[user_agent_key] == config.userAgent

        # modified config user-agent
        config.userAgent = 'Python Relayr API Super Special Testing Agent'
        headers = Api().get_server_status()
        assert headers[user_agent_key] == config.userAgent

        t.stop()

        config.relayrAPI = previous



# -*- coding: utf-8 -*-

"""
Data Connection Classes

This module provide connection classes for accessing device data.
"""

import time
import threading

from Pubnub import Pubnub

from relayr import config


class PubnubDataConnection(threading.Thread):
    "A connection to a PubNub data hub running on its own thread."
    
    def __init__(self, callback, credentials):
        """
        Opens a PubNub connection with a callback and a credentials dict.

        :param callback: A callable to be called with two arguments: 
            message_content and channel_name.
        :type callback: A function or object implementing the ``__call__`` method.
        :param credentials: A set of key/value pairs to be passed to PubNub.
        :type credentials: dict
        """

        super(PubnubDataConnection, self).__init__()
        self._stop_event = threading.Event()
        self.callback = callback
        self.credentials = credentials
        self.channel = credentials['channel']

        self.hub = Pubnub(
            publish_key=credentials.get('publishKey', None),
            subscribe_key=credentials.get('subscribeKey', None),
            cipher_key=credentials.get('cipherKey', None),
            auth_key=credentials.get('authKey', None),
            secret_key=None,
            ssl_on=True
        )

        self.setDaemon(True)

    def run(self):
        """Thread method, called implicitly after starting the thread."""

        self.subscribe(self.channel, self.callback)
        while not self._stop_event.is_set():
            time.sleep(1)

    def stop(self):
        """Mark the connection/thread for being stopped."""

        self._stop_event.set()
        self.unsubscribe(self.channel)

    def subscribe(self, channel_name, callback):
        """
        Subscribes a callable to a channel with the given name.

        :param channel_name: The channel name.
        :type channel_name: string
        :param callback: The callable to be called with two arguments: 
            message_content and channel_name.
        :type callback: A function or object implementing the ``__call__`` method.
        """
        
        self.hub.subscribe(channel_name, callback)

    def unsubscribe(self, channel_name):
        """
        Unsubscribes from a channel with given name.

        :param channel_name: the channel name
        :type channel_name: string
        """
        
        self.hub.unsubscribe(channel_name)


class MqttDataConnection(object):
    "..."
    
    def __init__(self, callback, credentials):
        pass

    def subscribe(self, channel_name, callback):
        raise NotImplementedError

    def unsubscribe(self, channel_name):
        raise NotImplementedError


if config.dataConnectionHubName == 'PubNub':
    Connection = PubnubDataConnection
elif config.dataConnectionHubName == 'MQTT':
    Connection = MqttDataConnection



# -*- coding: utf-8 -*-

"""
This module contains tests of the Relayr restful HTTP API client.

The tests are written in plain Python code, without using the unittest frame-
work, so you can run them via py.test (an alternative to unittest! As the
functions are also mostly fully self-contained you can copy/paste/run individual
test method bodies in an interactive Python session! And where the test methods
rely on some fixture parameter the values they provide can be often easily
replaced in copied code.

By convention, test fixtures are Python files with a ``fixture_`` prefix and
provide values used inside the tests like e.g. a user email address or access
token. The fixture loaders are contained in ``conftest.py`` where they will be
discovered by pytest.

Attention: The values in these fixture files need to be set properly before
running the tests!
"""

import pytest


class TestDeviceBookmarks(object):
    "Test device bookmarks."

    def test_get_user_bookmarked_devices(self, fix_registered):
        "Test get user's bookmarked devices."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        user = c.get_user()
        b_devs = user.get_bookmarked_devices()
        assert type(list(b_devs)) == list

    def test_post_user_devices_bookmarks(self, fix_registered):
        "Test get user's bookmarked devices."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        user = c.get_user()


class TestAPI(object):
    "Test low-level access to API endpoints."

    def test_validate_registered_user(self, fix_registered):
        "Test validate an existing user email address."
        from requests import get
        context = fix_registered.testset1
        url = '{relayrAPI}/users/validate?email={userEmail}'.format(**context)
        resp = get(url)
        assert resp.status_code == 200
        j = resp.json()
        assert resp.json()['exists'] == True

    def test_get_appdev_token(self, fix_registered):
        "Test get appdev token."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        app = c.get_app()
        res = c.api.get_oauth2_appdev_token(app.id)
        assert res['token'] == token

    @pytest.mark.skip(reason="unfinished") ###################
    def test_post_oauth2_token(self, fix_registered):
        "Test post oauth2 token."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        # res = c.api.post_oauth2_token('clientID', 'clientSecret', 'code', 'redirectURI')

    def test_connected_user(self, fix_registered):
        "Test get info about the currently connected user."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        info = c.api.get_oauth2_user_info()
        for key in 'name email id'.split():
            assert key in info

    def test_connected_app(self, fix_registered):
        "Test get info about the currently connected app."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        info = c.api.get_oauth2_app_info()
        for key in 'name description id'.split():
            assert key in info

    def test1_get_token(self, fix_registered):
        "Test get current token and compare with the one given to connect."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        app_id = c.api.get_oauth2_app_info()['id']
        info = c.api.get_oauth2_appdev_token(app_id)
        for key in 'token expiryDate'.split():
            assert key in info
        assert token == info['token']

    def _test_update_user(self, fix_registered):
        "Test update user attribute, here: name."
        from relayr import Client
        ## TODO: check if a regular user can change his own attributes...
        token = fix_registered.testset1['token']
        c = Client(token=token)
        info = c.api.get_oauth2_user_info()
        c.api.patch_user(info['id'], name='dcg')

    def test_get_transmitters(self, fix_registered):
        "Test get user's transmitters."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        user_info = c.api.get_oauth2_user_info()
        trans_info = c.api.get_user_transmitters(user_info['id'])
        for trans in trans_info:
            for key in 'owner name id'.split(): # plus others...
                assert key in trans

    def test_apps_of_connected_user(self, fix_registered):
        "Get apps of connected user."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        apps = c.api.get_user_apps(c.api.get_oauth2_user_info()['id'])
        assert len(apps) > 0
        assert type(apps[0]) == dict

    def test_public_device_credentials(self, fix_registered):
        "Test get credentials for subscribing to a public device."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        deviceID = fix_registered.testset1['deviceID']
        creds = c.api.post_devices_public_subscription(deviceID)
        for key in 'authKey subscribeKey cipherKey channel'.split():
            assert key in creds

    def test_get_wunderbar(self, fix_registered):
        "Test get info about the user's registered wunderbar device."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        user_info = c.api.get_oauth2_user_info()
        wb_dev_info = c.api.post_user_wunderbar(user_info['id'])
        assert len(wb_dev_info) == 7 # 6 sensors plus one transmitter
        owner_ids = [dev['owner'] for dev in wb_dev_info.values()]
        assert len(set(owner_ids)) == 1
        assert user_info['id'] == owner_ids[0]
        for typ, info in wb_dev_info.items():
            for key in 'owner name id'.split(): # plus others...
                assert key in info


class TestClient(object):
    "Test high-level access via API client."

    def test_public_resources(self, fix_registered):
        "Get public resources."
        from relayr import Client
        from relayr.resources import App, Publisher, Device, DeviceModel
        token = fix_registered.testset1['token']
        c = Client(token=token)
        assert next(c.get_public_apps()).__class__ == App
        assert next(c.get_public_devices()).__class__ == Device
        assert next(c.get_public_device_models()).__class__ == DeviceModel
        assert next(c.get_public_publishers()).__class__ == Publisher

    def test_users_resources(self, fix_registered):
        "Test get user's resources."
        from relayr import Client
        from relayr.resources import User, Publisher, App, Device
        token = fix_registered.testset1['token']
        c = Client(token=token)
        usr = c.get_user()
        assert usr.__class__ == User
        assert c.get_app().__class__ == App
        assert next(usr.get_apps()).__class__ == App
        assert next(usr.get_devices()).__class__ == Device
        assert next(usr.get_publishers()).__class__ == Publisher

    def test_user_name(self, fix_registered):
        "Test get name attribute of User object for connected user."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        usr = c.get_user()
        assert usr.name == fix_registered.testset1['userName']

    def test4(self, fix_registered):
        "Test get publishers for connected user."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        usr = c.get_user()
        # pubs = list(usr.get_publishers()) # The supplied authentication is not authorized to access this resource.

    def test_access_device(self, fix_registered):
        "Test access a public device."
        from relayr import Client
        from relayr.resources import Device
        token = fix_registered.testset1['token']
        deviceID = fix_registered.testset1['deviceID']
        c = Client(token=token)
        dev = Device(id=deviceID)

    def test_send_command_device(self, fix_registered):
        "Test send a command to a device."
        from relayr import Client
        from relayr.resources import Device
        token = fix_registered.testset1['token']
        c = Client(token=token)
        deviceID = fix_registered.testset1['deviceID']
        dev = Device(id=deviceID, client=c)
        dev.get_info()
        dev.send_command('led', {'cmd': 1})
        # Eye-ball test, since reading the LED status is not yet implemented
        # in the firmware.

    # @pytest.xfail("no api method?")
    def test_publish_device(self):
        "Test make a device public/private."

    def test_get_wunderbar_devices(self, fix_registered):
        "Test get user's registered wunderbar devices and master."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        usr = c.get_user()
        items = list(usr.register_wunderbar())
        assert len(items) == 7


class TestTransmitters(object):
    "Tests on transmitters."

    def test_get_transmitter_devices(self, fix_registered):
        "Test getting devices of a known transmitter."
        from relayr import Client
        from relayr.resources import Transmitter
        token = fix_registered.testset1['token']
        transmitterID = fix_registered.testset1['transmitterID']
        c = Client(token=token)
        t = Transmitter(id=transmitterID, client=c)
        t.get_info()
        assert t.name == fix_registered.testset1['transmitterName']
        devs = t.get_connected_devices()
        assert len(list(devs)) == 6


class TestApps(object):
    "Tests on apps."

    def test_app_info(self, fix_registered):
        "Test getting info for the app."
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        app = c.get_app()
        assert hasattr(app, 'id')

    def test_app_update(self, fix_registered):
        "Test update info for some app."
        # This is actually done twice for making this an idempotent operation.
        from relayr import Client
        token = fix_registered.testset1['token']
        c = Client(token=token)
        app = c.get_app()
        initial_app_name = app.name
        # update name to something silly
        silly_name = 'HolyWhat?'
        app.update(name=silly_name)
        assert app.name == silly_name
        # update name to previous name
        app.update(name=initial_app_name)        
        assert app.name == initial_app_name


class TestWunderbar(object):
    "Tests accessing a wunderbar."

    def test_find_wunderbar_devices(self, fix_registered):
        "Test get all devices on a Wunderbar."
        from relayr import Client
        from relayr.resources import Device, Transmitter
        token = fix_registered.testset1['token']
        c = Client(token=token)
        usr = c.get_user()
        assert usr.name == fix_registered.testset1['userName']
        devs = usr.register_wunderbar()
        for d in devs:
            assert d.__class__ in (Device, Transmitter)
            assert hasattr(d, 'id')


class TestPublishers(object):
    "Tests on publishers."

    def test_pub_get_apps(self, fix_registered):
        "Test getting apps for some publisher."
        from relayr import Client
        from relayr.resources import Publisher
        publisherID = fix_registered.testset1['publisherID']
        pub = Publisher(id=publisherID, client=Client())
        apps = pub.get_apps()
        app_list = list(apps)
        assert len(app_list) >= 1
        assert hasattr(app_list[0], 'id')

    def test_pub_get_apps_extended(self, fix_registered):
        "Test getting extended apps for some publisher."
        from relayr import Client
        from relayr.resources import Publisher
        token = fix_registered.testset1['token']
        c = Client(token=token)
        publisherID = fix_registered.testset1['publisherID']
        pub = Publisher(id=publisherID, client=c)
        apps = pub.get_apps(extended=True)
        app_list = list(apps)
        assert len(app_list) >= 1
        assert hasattr(app_list[0], 'clientId')



"""

"""

import platform

from relayr import config
from relayr.api import Api
from relayr.version import __version__
from relayr.exceptions import RelayrApiException
from relayr.resources import User, App, Device, DeviceModel, Transmitter, Publisher


class Client(object):
    """
    A client providing a higher level interface to the relayr cloud platform.

    Example:

    .. code-block:: python

        c = Client(token='...')
        info = c.get_oauth_user_info()
        usr = User(info['id'], client=c)
        devs = usr.get_devices()
        d = next(devs)
        apps = usr.get_apps()
    """
    def __init__(self, token=None):
        """
        :arg token: A token generated on the relayr site for the combination of
            a user and an application.
        :type token: A string.
        """

        self.api = Api(token=token)

    def get_public_apps(self):
        """
        Returns a generator for all apps on the relayr platform.

        A generator is returned since the called API method always
        returns the entire results list and not a paginated one.
        

        :rtype: A generator for :py:class:`relayr.resources.App` objects.

        .. code-block:: python

            c = Client(token='...')
            apps = c.get_public_apps()
            for app in apps:
                print('%s %s' % (app.id, app.name))
        """

        for app in self.api.get_public_apps():
            a = App(app['id'], client=self)
            a.get_info()
            yield a

    def get_public_publishers(self):
        """
        Returns a generator for all publishers on the relayr platform.

        A generator is returned since the called API method always
        returns the entire results list and not a paginated one.
        

        :rtype: A generator for :py:class:`relayr.resources.Publisher` objects.
        """

        for pub in self.api.get_public_publishers():
            p = Publisher(pub['id'], client=self)
            for k in pub:
                setattr(p, k, pub[k])
            # p.get_info()
            yield p

    def get_public_devices(self, meaning=''):
        """
        Returns a generator for all devices on the relayr platform.

        A generator is returned since the called API method always
        returns the entire results list and not a paginated one.
        

        :arg meaning: The *meaning* (type) of the desired devices.
        :type meaning: string
        :rtype: A generator for :py:class:`relayr.resources.Device` objects.
        """

        for dev in self.api.get_public_devices(meaning=meaning):
            d = Device(dev['id'], client=self)
            d.get_info()
            yield d

    def get_public_device_models(self):
        """
        Returns a generator for all device models on the relayr platform.

        A generator is returned since the called API method always
        returns the entire results list and not a paginated one.
        

        :rtype: A generator for :py:class:`relayr.resources.DeviceModel` objects.
        """

        for dm in self.api.get_public_device_models():
            d = DeviceModel(dm['id'], client=self)
            d.get_info()
            yield d

    def get_public_device_model_meanings(self):
        """
        Returns a generator for all device models' meanings on the relayr platform.

        A device model meaning is a simple dictionary with a ``key`` and ``value``
        field.

        A generator is returned since the called API method always
        returns the entire results list and not a paginated one.
        
        
        :rtype: A device model meaning (as a dictionary) generator.

        .. code-block:: python

            {'key': 'humidity', 'value': 'humidity'}
        """

        for dmm in self.api.get_public_device_model_meanings():
            yield dmm

    def get_user(self):
        """
        Returns the relayr user owning the API client.

        :rtype: A :py:class:`relayr.resources.User` object.
        """
        info = self.api.get_oauth2_user_info()
        usr = User(info['id'], client=self)
        for k in info:
            setattr(usr, k, info[k])
        return usr

    def get_app(self):
        """
        Returns the relayr application of the API client.

        :rtype: A :py:class:`relayr.resources.App` object.
        """
        info = self.api.get_oauth2_app_info()
        app = App(info['id'], client=self)
        app.get_info()
        return app

    def get_device(self, id):
        """
        Returns the device with the specified ID.

        :arg id: the unique ID for the desired device.
        :type id: string
        :rtype: A :py:class:`relayr.resources.Device` object.
        """
        return Device(id=id, client=self)



"""
:copyright: Copyright 2014 by the relayr.io team, see AUTHORS.txt.
:license: MIT, see LICENSE.txt for details.
"""

from relayr.client import Client, Api


def show_docs(url=''):
    "Open online documentation with the default webbrowser in a new window."

    import webbrowser
    url = url or 'http://relayr.readthedocs.org/'
    webbrowser.open_new(url)


## TODO: Start IPython notebook in the background and open given file (?).
def show_tutorial(path=''):
    "Open IPython notebook with given path."

    # see https://gist.github.com/minrk/2620876



# -*- coding: utf-8 -*-

"""
This module contains abstractions for relayr API resources.

Resources may be entities such as users, publishers, applications, 
devices, device models and transmitters.
"""


from relayr import exceptions
from relayr.dataconnection import Connection


class User(object):
    "A Relayr user."

    def __init__(self, id=None, client=None):
        self.id = id
        self.client = client

    def __repr__(self):
        return "%s(id=%r)" % (self.__class__.__name__, self.id)

    def get_publishers(self):
        "Return a generator of the publishers of the user."

        for pub_json in self.client.api.get_user_publishers(self.id):
            p = Publisher(pub_json['id'], client=self.client)
            for k in pub_json:
                setattr(p, k, pub_json[k])
            yield p

    def get_apps(self):
        "Returns a generator of the apps of the user."

        for app_json in self.client.api.get_user_apps(self.id):
            ## TODO: change 'app' field to 'id' in API?
            app = App(app_json['app'], client=self.client)
            app.get_info()
            yield app

    def get_transmitters(self):
        "Returns a generator of the transmitters of the user."

        for trans_json in self.client.api.get_user_transmitters(self.id):
            trans = Transmitter(trans_json['id'], client=self.client)
            trans.get_info()
            yield trans

    def get_devices(self):
        "Returns a generator of the devices of the user."

        for dev_json in self.client.api.get_user_devices(self.id):
            dev = Device(dev_json['id'], client=self.client)
            dev.get_info()
            yield dev

    def connect_device(self, app, device, callback):
        "Opens and returns a connection to the data provider."

        creds = self.client.api.post_devices_subscription(app.id, device.id)
        return Connection(callback, creds)

    def connect_public_device(self, device, callback):
        "Opens and returns a connection to the data provider."

        creds = self.client.api.post_devices_public_subscription(device.id)
        return Connection(callback, creds)

    def disconnect_device(self, id):
        # There is no disconnect in the API...
        raise NotImplementedError

    def update(self, name=None, email=None):
        res = self.client.api.patch_user(self.id, name=name, email=email)
        for k in res:
            setattr(self, k, res[k])
        return self

    ## TODO: rename to 'registered_wunderbar_devices'?
    def register_wunderbar(self):
        """
        Returns registered Wunderbar devices (master and sensor modules).
    
        :rtype: A generator over the registered devices and one transmitter.
        """    
        res = self.client.api.post_user_wunderbar(self.id)
        for k, v in res.items():
            if 'model' in v:
                item = Device(res[k]['id'], client=self.client)
                item.get_info()
            else:
                item = Transmitter(res[k]['id'], client=self.client)
                item.get_info()
            yield item

    def remove_wunderbar(self):
        """
        Removes all Wunderbars associated with the user.
        """
        res = self.client.api.post_users_destroy(self.id)
        return res

    def get_bookmarked_devices(self):
        """
        Retrieves a list of bookmarked devices.

        :rtype: list of device objects
        """
        res = self.client.api.get_user_devices_bookmarks(self.id)
        for dev in res:
            d = Device(dev['id'], client=self.client)
            for k, v in dev.items():
                setattr(d, k, v)
            d.get_info()
            yield d

    def bookmark_device(self, device):
        res = self.client.api.post_user_devices_bookmark(self.id, device.id)

    def delete_device_bookmark(self, device):
        res = self.client.api.delete_user_devices_bookmark(self.id, device.id)
        return res


class Publisher(object):
    """
    A relayr publisher.

    A publisher has a few attributes, which can be updated. It can be
    registered to and deleted from the relayr platform. It can list all
    applications it has published on the relayr platform.
    """

    def __init__(self, id=None, client=None):
        self.id = id
        self.client = client

    def __repr__(self):
        return "%s(id=%r)" % (self.__class__.__name__, self.id)

    def get_apps(self, extended=False):
        """
        Get list of apps for this publisher.

        If the optional parameter ``extended`` is ``False`` (default) the 
        response will contain only the fields ``id``, ``name`` and 
        ``description``. If it is ``True`` it will contain additional 
        fields: ``publisher``, ``clientId``, ``clientSecret`` and
        ``redirectUri``.

        :param extended: Flag indicating if the info should be extended.
        :type extended: booloean
        :rtype: A list of :py:class:`relayr.resources.App` objects.
        """

        func = self.client.api.get_publisher_apps
        if extended:
            func = self.client.api.get_publisher_apps_extended
        res = func(self.id)
        for a in res:
            app = App(a['id'], client=self.client)
            app.get_info(extended=extended)
            yield app


    def update(self, name=None):
        """
        Updates certain information fields of the publisher's.
        
        :param name: the user email to be set
        :type name: string
        """
        res = self.api.patch_publisher(self.id, name=name)
        for k in res:
            setattr(self, k, res[k])
        return self

    def register(self, name, id, publisher):
        """
        Adds the publisher to the relayr platform.

        :param name: the publisher name to be set
        :type name: string
        :param id: the publisher UID to be set
        :type id: string
        """
        raise NotImplementedError

    def delete(self):
        """
        Deletes the publisher from the relayr platform.
        """
        res = self.api.delete_publisher(self.id)


class App(object):
    """
    A relayr application.
    
    An application has a few attributes, which can be updated. It can be
    registered to and deleted from the relayr platform. it can be connected 
    to and disconnected from devices.
    """
    
    def __init__(self, id=None, client=None):
        self.id = id
        self.client = client

    def __repr__(self):
        return "%s(id=%r)" % (self.__class__.__name__, self.id)

    def get_info(self, extended=False):
        """
        Get application info.
        
        If the optional parameter ``extended`` is ``False`` (default) the 
        result will contain only the fields ``id``, ``name`` and 
        ``description``. If it is ``True`` it will contain additional 
        fields: ``publisher``, ``clientId``, ``clientSecret`` and
        ``redirectUri``.
        
        :param extended: flag indicating if the info should be extended
        :type extended: booloean
        :rtype: A dict with certain fields.
        """

        func = self.client.api.get_app_info
        if extended:
            func = self.client.api.get_app_info_extended
        res = func(self.id)
        for k in res:
            setattr(self, k, res[k])
        return self

    def update(self, description=None, name=None, redirectUri=None):
        """
        Updates certain fields in the application's description.
        
        :param description: the user name to be set
        :type description: string
        :param name: the user email to be set
        :type name: string
        :param redirectUri: the redirect URI to be set
        :type redirectUri: string
        """
        res = self.client.api.patch_app(self.id, description=description,
            name=name, redirectUri=redirectUri)
        for k in res:
            setattr(self, k, res[k])
        return self

    def delete(self):
        """
        Deletes the app from the relayr platform.
        """
        res = self.api.delete_publisher(self.id)

    def register(self, name, publisher):
        """
        Adds the app to the relayr platform.

        :param name: the app name to be set
        :type name: string
        :param publisher: the publisher to be set
        :type publisher: string(?)
        """
        raise NotImplementedError

    def connect_to_device(self, device):
        """
        Connects the app to a device.
        
        Data reception credentials are returned as part of the response.

        See also the Device.connect_to_device() method...

        :param device: the device (name) to be connected
        :type device: string(?)
        """
        raise NotImplementedError

    def disconnect_from_device(self, device):
        """
        Disonnects the app from a device.

        See also the Device.disconnect_from_app() method...

        :param device: the device (name) to be disconnected from
        :type device: string(?)
        """
        raise NotImplementedError


class Device(object):
    """
    A relayr device.
    """

    def __init__(self, id=None, client=None):
        self.id = id
        self.client = client

    def __repr__(self):
        return "%s(id=%r)" % (self.__class__.__name__, self.id)

    def get_info(self):
        """
        Retrieves device info and stores it as instance attributes.

        :rtype: self.
        """

        res = self.client.api.get_device(self.id)
        for k in res:
            if k == 'model':
                self.model = DeviceModel(res[k]['id'], client=self.client)
                self.model.get_info()
            else:
                setattr(self, k, res[k])
        return self

    def update(self, description=None, name=None, model=None, public=None):
        """
        Updates certain fields in the device description.
        
        :param description: the description to be set
        :type description: string
        :param name: the user name to be set
        :type name: string
        :param model: the device model to be set
        :type name: string?
        :param public: a flag for making the device public
        :type redirectUri: boolean
        """
        res = self.client.api.patch_device(self.id, deviceDescription=description,
            deviceName=name, deviceModel=model, isDevicePublic=public)
        for k in res:
            setattr(self, k, res[k])
        return self

    def get_connected_apps(self):
        """
        Retrieves all apps connected to the device.
        
        :rtype: A list of apps.
        """

        res = self.client.api.get_device_apps(self.id)
        return res

    def connect_to_app(self, app):
        """
        Connects the device to an app.
        
        Data reception credentials are returned as part of the response.

        See also the App.connect_to_device() method...

        :param app: the app (name) to be connected
        :type app: string(?)
        """
        res = self.client.api.post_app_device(app.id, self.id)
        return res

    def disconnect_from_app(self, app):
        """
        Disconnects the device from an app.

        See also the App.disconnect_from_device() method...

        :param app: the app (name) to be disconnected from
        :type app: string(?)
        """
        res = self.client.api.delete_app_device(app.id, self.id)
        return res

    def connect_to_device(self, appID, id, callback):
        """
        Subscribes a user to a device.
        """
        creds = self.client.api.post_devices_subscription(appID, self.id)
        return Connection(callback, creds)

    def connect_to_public_device(self, id, callback):
        """
        Subscribes a user to a public device.

        :param id: the device's UID
        :type id: string
        """

        creds = self.client.api.post_devices_public_subscription(self.id)
        return Connection(callback, creds)

    def send_command(self, command, data):
        """
        Sends a command to the device.

        :param command: the command to be sent
        :type command: string
        :param data: the command data to be sent
        :type command: dict
        """
        
        res = self.client.api.post_device_command(self.id, command, data)
        return res

    def delete(self):
        """
        Deletes the device from the relayr platform.

        :type command: self
        """
        
        res = self.client.api.delete_device(self.id)
        return self

    def switch_led_on(self, bool=True):
        """
        Switches on device's LED for ca. 10 seconds or switches it off.

        :param bool: the desired state, on if True (default), off if False
        :type bool: boolean
        :type command: self
        """
        self.send_command('led', {'cmd': int(bool)})
        return self

class DeviceModel(object):
    """
    relayr device model.
    """
    
    def __init__(self, id=None, client=None):
        self.id = id
        self.client = client

    def __repr__(self):
        return "%s(id=%r)" % (self.__class__.__name__, self.id)

    def get_info(self):
        """
        Returns device model info and stores it as instance attributes.
        
        :rtype: self.
        """
        res = self.client.api.get_device_model(self.id)
        for k, v in res.items():
            setattr(self, k, v)
        return self


class Transmitter(object):
    "A relayr transmitter, The Master Module, for example."
    
    def __init__(self, id=None, client=None):
        self.id = id
        self.client = client

    def __repr__(self):
        args = (self.__class__.__name__, self.id)
        return "%s(id=%r)" % args

    def get_info(self):
        """
        Retrieves transmitter info.
        """
        res = self.client.api.get_transmitter(self.id)
        for k, v in res.items():
            setattr(self, k, v)
        return self

    def delete(self):
        """
        Deletes the transmitter from the relayr platform.

        :type command: self
        """
        
        res = self.client.api.delete_transmitter(self.id)
        return self

    def update(self, name=None):
        """
        Updates transmitter info.
        """
        res = self.client.api.patch_transmitter(self.id, name=name)
        for k, v in res.items():
            setattr(self, k, v)
        return self

    def get_connected_devices(self):
        """
        Returns a list of devices connected to the specific transmitter.
        
        :rtype: A list of devices.
        """
        res = self.client.api.get_transmitter_devices(self.id)
        for d in res:
            dev = Device(d['id'], client=self.client)
            dev.get_info()
            yield dev



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Scan Wunderbar devices (BLE) and show what we find about them.

This is a use-case of the very experimental ``ble`` module. 

Todo:

- add command-line args to make this more flexible, maybe

Example:

$ python scan_wunderbars.py
Discovered device: F1:42:E6:63:20:A2 WunderbarMIC 
  characteristic uuid: 00002a00-0000-1000-8000-00805f9b34fb
    name handle: 0x0002, data: 02 03 00 00 2a , str: '\x02\x03\x00\x00*'
    value handle: 0x0003, data: 57 75 6e 64 65 72 62 61 72 4d 49 43 , str: 'WunderbarMIC'
  characteristic uuid: 00002a01-0000-1000-8000-00805f9b34fb
    name handle: 0x0004, data: 02 05 00 01 2a , str: '\x02\x05\x00\x01*'
    value handle: 0x0005, data: 00 02 , str: '\x00\x02'
  characteristic uuid: 00002a04-0000-1000-8000-00805f9b34fb
    name handle: 0x0006, data: 02 07 00 04 2a , str: '\x02\x07\x00\x04*'
    value handle: 0x0007, data: b0 00 b0 00 05 00 2c 01 , str: '\xb0\x00\xb0\x00\x05\x00,\x01'
  characteristic uuid: 00002a05-0000-1000-8000-00805f9b34fb
    name handle: 0x0009, data: 20 0a 00 05 2a , str: ' \n\x00\x05*'
    value handle: 0x000a, data: None, str: None
  characteristic uuid: 00002010-0000-1000-8000-00805f9b34fb
    name handle: 0x000d, data: 02 0e 00 10 20 , str: '\x02\x0e\x00\x10 '
    value handle: 0x000e, data: None, str: None
  characteristic uuid: 00002011-0000-1000-8000-00805f9b34fb
    name handle: 0x0010, data: 0a 11 00 11 20 , str: '\n\x11\x00\x11 '
    value handle: 0x0011, data: None, str: None
  characteristic uuid: 00002012-0000-1000-8000-00805f9b34fb
    name handle: 0x0013, data: 0a 14 00 12 20 , str: '\n\x14\x00\x12 '
    value handle: 0x0014, data: None, str: None
  characteristic uuid: 00002013-0000-1000-8000-00805f9b34fb
    name handle: 0x0016, data: 08 17 00 13 20 , str: '\x08\x17\x00\x13 '
    value handle: 0x0017, data: None, str: None
  characteristic uuid: 00002014-0000-1000-8000-00805f9b34fb
    name handle: 0x0019, data: 0a 1a 00 14 20 , str: '\n\x1a\x00\x14 '
    value handle: 0x001a, data: None, str: None
  characteristic uuid: 00002016-0000-1000-8000-00805f9b34fb
    name handle: 0x001c, data: 32 1d 00 16 20 , str: '2\x1d\x00\x16 '
    value handle: 0x001d, data: None, str: None
  characteristic uuid: 00002a29-0000-1000-8000-00805f9b34fb
    name handle: 0x0021, data: 02 22 00 29 2a , str: '\x02"\x00)*'
    value handle: 0x0022, data: 52 65 6c 61 79 72 , str: 'Relayr'
  characteristic uuid: 00002a27-0000-1000-8000-00805f9b34fb
    name handle: 0x0023, data: 02 24 00 27 2a , str: "\x02$\x00'*"
    value handle: 0x0024, data: 31 2e 30 2e 32 , str: '1.0.2'
  characteristic uuid: 00002a26-0000-1000-8000-00805f9b34fb
    name handle: 0x0025, data: 02 26 00 26 2a , str: '\x02&\x00&*'
    value handle: 0x0026, data: 31 2e 30 2e 30 , str: '1.0.0'
  characteristic uuid: 00002a19-0000-1000-8000-00805f9b34fb
    name handle: 0x0028, data: 12 29 00 19 2a , str: '\x12)\x00\x19*'
    value handle: 0x0029, data: 64 , str: 'd'
  device name: WunderbarMIC
  battery level: 100
...
"""


from relayr.ble import scan_ble_devices, GattDevice, data2str


def scan_wunderbars():
    """
    Scan Wunderbar devices and show what we find about them.
    """
    
    devices = scan_ble_devices(name_filter='Wunderbar.*')
    for dev in devices:
        print("Discovered device: %(addr)s %(name)s " % dev)
        d = GattDevice(dev['addr'])
        chars = d.characteristics()
        for char in chars:
            if len(char['_uuid']) == 4:
                print('  characteristic uuid: %s' % char['uuid'])
                h = char['name_handle']
                dat = d.char_read_hnd(h)
                args = (h, dat, repr(data2str(dat)))
                print('    name handle: %s, data: %s, str: %s' % args)
                h = char['value_handle']
                dat = d.char_read_hnd(h)
                args = (h, dat, repr(data2str(dat)))
                print('    value handle: %s, data: %s, str: %s' % args)
        print('  device name: %s' % d.read_device_name())
        print('  battery level: %s' % d.read_battery_level())
        print('')


if __name__ == "__main__":
    scan_wunderbars()



# A set of variables used for testing the relayr API.
# Attention: You must provide your own eaningful values 
# before you can run large parts of the test suite!

# Get 'token' on relayr.io!

# Remember that tokens expire after some time, check the documentation!

testset1 = {
    'userID': 'd320f267-4bba-433d-a59a-2e9fc881df8f', 
    'userName': 'dinugherman', 
    'userEmail': 'gherman@darwin.in-berlin.de',
    'relayrAPI': 'https://api.relayr.io',
    'redirectURI': 'https://relayr.io',
    'response_type': 'code', # or 'token'
    'clientId': 'CzmPlGbCtUI8tAwvlSJkW8IcRjazCp6u',
    'clientSecret': '05UlzL2r3zFf7aeAPmZMOF2wnBHdd5Ue',
    'scope': 'access-own-user-info',
    'appID': 'c6941a2b-3b2f-404f-962d-cd88328eff26',
    'appName': 'DGApp0',
    'publisherID': 'f4002d7d-f6e9-46b8-b79a-dbb6c435f282',
    'deviceID': 'a3386628-61a8-4a80-9a42-1bd3c64c46a5',
    'description': "Dinu's first Wunderbar app",
    'useragent': 'io.relayr.postman/0.1.0',
    'token': 'KkD_V4TnjmN7Mw_zceRhZuNf1GCG_N9X',
    'transmitterID': '4f3b23be-53b7-4b1d-8da7-379a0b81eeba',
    'transmitterName': 'DCG08 My Wunderbar',
}

testset3 = {
    'PUBLISH_KEY': 'pub-c-26992c4a-5a1c-4ae1-ba04-bb01f87ec12b',
    'SUBSCRIBE_KEY': 'sub-c-b702a884-43fe-11e4-a249-02ee2ddab7fe',
    'channel': 'test-channel',
    'message': 'hi there 11'
}

testset4 = {
    'bluetoothAddr': 'f1:42:e6:63:20:a2',
}



# -*- coding: utf-8 -*-

"""
This module contains tests of the Relayr restful HTTP API client.

The tests are written in plain Python code, without using the unittest frame-
work, so you can run them via py.test (an alternative to unittest! As the
functions are also mostly fully self-contained you can copy/paste/run individual
test method bodies in an interactive Python session! And where the test methods
rely on some fixture parameter the values they provide can be often easily
replaced in copied code.

By convention, test fixtures are Python files with a ``fixture_`` prefix and
provide values used inside the tests like e.g. a user email address or access
token. The fixture loaders are contained in ``conftest.py`` where they will be
discovered by pytest.

Attention: The values in these fixture files need to be set properly before
running the tests!
"""

import pytest


class TestInstallation(object):
    "Test installation aspects."

    def test_imports(self):
        "Test importing various bits."
        from relayr.version import __version__
        from relayr import Client
        from relayr.resources import User, Publisher, App, Transmitter, \
            Device, DeviceModel
        from relayr.exceptions import RelayrApiException
        from relayr import Api
        from relayr import config


class TestRawAPI(object):
    "Test raw HTTP access to some API endpoints."

    def test_raw_server_status(self, fix_anonymous):
        "Test server status."
        from requests import get
        relayrAPI = fix_anonymous.testset0['relayrAPI']
        resp = get('%s/server-status' % relayrAPI)
        assert resp.json() == {'database': 'ok'}

    def test_raw_non_existing_endpoint(self, fix_anonymous):
        "Test accessing a non-existing API endpoint."
        from requests import get
        relayrAPI = fix_anonymous.testset0['relayrAPI']
        resp = get('%s/this-should-not-exist' % relayrAPI)
        assert resp.status_code == 404

    def test_raw_all_publishers(self, fix_anonymous):
        "Test accessing list of all public publishers."
        # More or less the same for othere resources like devices, etc. 
        # with partly different expected keys:
        #   /device-models/meanings: key, value
        #   /device-models: id, readings, firmwareVersions, manufacturer, name
        #   /devices/public: model, firmwareVersion, id, name 
        #   /apps: id, name 
        from requests import get
        relayrAPI = fix_anonymous.testset0['relayrAPI']
        resp = get('%s/publishers' % relayrAPI)
        assert resp.status_code == 200
        publishers = resp.json()
        assert type(publishers) == list
        expected_keys = set(['id', 'name', 'owner'])
        for p in publishers:
            assert expected_keys.issubset(set(p.keys())) == True


class TestRequests(object):
    "Test performing requests."

    def test_unknown_endpoint(self, fix_anonymous):
        "Test accessing an unknown API endpoint."
        from relayr import Api
        from relayr.exceptions import RelayrApiException
        with pytest.raises(RelayrApiException) as excinfo:
            url = '%s/this-should-not-exist' % fix_anonymous.testset0['relayrAPI']
            resp = Api().perform_request('GET', url)
        assert str(excinfo.value).startswith('URL could not be routed.')

    def test_unknown_endpoint_data(self, fix_anonymous):
        "Test accessing an unknown API endpoint with body data."
        from relayr import Api
        from relayr.exceptions import RelayrApiException
        with pytest.raises(RelayrApiException) as excinfo:
            url = '%s/this-should-not-exist' % fix_anonymous.testset0['relayrAPI']
            resp = Api().perform_request('POST', url, data={'foo': 42, 'bar': 23})
        assert str(excinfo.value).startswith('URL could not be routed.')
        assert '--data "foo=42&bar=23"' in str(excinfo.value) or \
            '--data "bar=23&foo=42"' in str(excinfo.value)


class TestAPIAnonymous(object):
    "Test low-level access to API endpoints without a need for credentials."

    def test_server_status(self):
        "Test server status."
        from relayr import Client
        assert Client().api.get_server_status() == {u'database': u'ok'}

    def test_users_validate_non_existing(self, fix_anonymous):
        "Test validate non-existing user email."
        from relayr import Client
        userName = fix_anonymous.testset0['userEmail']
        assert Client().api.get_users_validate(userName) == {u'exists': False}

    def test_exception_missing_creds(self):
        "Test raising exception for missing credentials."
        from relayr import Client
        from relayr.exceptions import RelayrApiException
        with pytest.raises(RelayrApiException) as excinfo:
            info = Client().api.get_oauth2_user_info()
        prefix = 'The supplied authentication is not authorized to access this resource.'
        assert str(excinfo.value).startswith(prefix)

    def test_list_all_publishers(self):
        from relayr import Client
        c = Client()
        expected_attrs = set(['id', 'name', 'owner'])
        for p in c.get_public_publishers():
            assert expected_attrs.issubset(set(dir(p))) == True



"""
Implementation of the relayr API as individual endpoints.

This module contains the API class with one method for each API endpoint.
All method names start with the HTTP method followed by the resource name used
in that endpoint e.g. ``post_user_app`` for the endpoint
``POST /users/<id>/apps/<id>`` with minor modifications.
"""

import os
import time
import json
import platform
import urllib
import warnings
import logging

import requests

from relayr import config
from relayr.version import __version__
from relayr.exceptions import RelayrApiException
from relayr.compat import urlencode


def create_logger(sender):
    "Create a logger for the requesting object."

    logger = logging.getLogger('Relayr API Client')
    logger.setLevel(logging.DEBUG)

    logfile = "{0}/relayr-api-{1}.log".format(config.LOG_DIR, id(sender))
    h = logging.FileHandler(logfile)
    # h = logging.RotatingFileHandler(logfile, 
    #     mode='a', maxBytes=2**14, backupCount=5, encoding=None, delay=0)

    # h.setLevel(logging.DEBUG)

    # create formatter and add it to the handler(s)
    fmt = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    formatter = logging.Formatter(fmt, '%Y-%m-%d %H:%M:%S.%f %Z%z')
    formatter.converter = time.gmtime
    h.setFormatter(formatter)

    # add the handler(s) to the logger
    logger.addHandler(h)

    return logger

def build_curl_call(method, url, data=None, headers=None):
    """
    Builds and returns a ``curl`` command for use on the command-line.

    (The data parameter should already be urlencoded...) ## FIXME
    """

    command = "curl -X {0} {1}".format(method.upper(), url)
    if headers:
        for k, v in headers.items():
            command += ' -H "{0}: {1}"'.format(k, v)
    if data:
        # Add body params (eg. --data "param1=value1&param2=value2")
        # command += ' --data "{0}"'.format(urlencode(data))
        command += ' --data "{0}"'.format(data)
    return command

class Api(object):
    """
    This class provides direct access to the relayr API endpoints.

    Examples:

    .. code-block:: python

        # Creates an anonymous client and calls simple API endpoints:
        from relayr.api import Api
        a = Api()
        assert a.get_server_status() == {'database': 'ok'}
        assert a.get_users_validate('god@in.heaven') == {'exists': False}
        assert a.get_public_device_model_meanings() > 0

    """
    def __init__(self, token=None):
        """
        :arg token: a token generated on the relayr site for a combination of 
            a relayr user and application.
        """

        self.token = token
        self.host = config.relayrAPI
        self.useragent = config.userAgent
        self.headers = {
            'User-Agent': self.useragent,
            'Content-Type': 'application/json'
        }
        if self.token:
            self.headers['Authorization'] = 'Bearer {0}'.format(self.token)

        if config.LOG:
            self.logger = create_logger(self)
            self.logger.info('started')

        # check if the API is available
        try:
            self.get_server_status()
        except:
            raise 

    def __del__(self):
        """Object destruction..."""
        if config.LOG:
            self.logger.info('terminated')

    def perform_request(self, method, url, data=None, headers=None):
        """
        Performs an API call and returns a JSON result as Python datastructure.

        Query parameters are expected in the ``url`` parameter.
        For returned status codes other than 2XX a ``RelayrApiException``
        is raised which contains the API call (method and URL) plus 
        a ``curl`` command replicating the API call for debugging reuse 
        on the command-line.
        """

        if config.LOG:
            command = build_curl_call(method, url, data, headers)
            self.logger.info("API request: " + command)

        urlencoded_data = None
        if data is not None:
            urlencoded_data = urlencode(data)
            data = json.dumps(data)
            try:
                data = data.encode('utf-8')
            except (UnicodeDecodeError, AttributeError):
                # bytes/str - no need to re-encode
                pass

        func = getattr(requests, method.lower())
        resp = func(url, data=data or '', headers=headers or {})
        resp.connection.close()

        if config.LOG:
            hd = dict(resp.headers.items())
            self.logger.info("API response headers: " + json.dumps(hd))
            self.logger.info("API response content: " + resp.content)

        status = resp.status_code
        if 200 <= status < 300:
            try:
                js = resp.json()
            except:
                js = None
                # raise ValueError('Invalid JSON code(?): %r' % resp.content)
                if config.DEBUG:
                    warnings.warn("Replaced suspicious API response (invalid JSON?) %r with 'null'!" % resp.content)
            return status, js
        else:
            args = (resp.json()['message'], method.upper(), url)
            msg = "{0} - {1} {2}".format(*args)
            command = build_curl_call(method, url, urlencoded_data, headers)
            msg = "%s - %s" % (msg, command)
            raise RelayrApiException(msg)

    # ..............................................................................
    # System
    # ..............................................................................

    def get_users_validate(self, userEmail):
        """
        Validate a user email address.
        
        :param userEmail: The user email address to be validated.
        :type userEmail: string
        :rtype: A dict with an ``exists`` field and a Boolean result value.
    
        Sample result::
    
            {"exists": True}
        """
        
        # https://api.relayr.io/users/validate?email=<userEmail>
        url = '{0}/users/validate?email={1}'.format(self.host, userEmail)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_server_status(self):
        """
        Queries server status.

        :rtype: A dict with certain fields describing the server status.

        Sample result::

            {"database": "ok"}
        """

        # https://api.relayr.io/server-status
        url = '{0}/server-status'.format(self.host)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def post_oauth2_token(self, clientID, clientSecret, code, redirectURI):
        """
        User:token from tmp code. (?)
        
        Generates a user token from supplied parameters
        """

        data = {
            "client_id": clientID,
            "client_secret": clientSecret,
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": redirectURI
        }
    
        # https://api.relayr.io/oauth2/token
        url = '{0}/oauth2/token'.format(self.host)
        _, data = self.perform_request('POST', url, data=data, headers=self.headers)
        return data

    def get_oauth2_appdev_token(self, appID):
        """
        Retrieves a token representing a specific relayr application and user.
        
        :param appID: The application's UUID.
        :type appID: string
        :rtype: A dict with fields describing the token.
    
        Sample result (anonymized token value)::
    
            {
                "token": "...",
                "expiryDate": "2014-10-08T10:14:07.789Z"
            }
        """
    
        # https://api.relayr.io/oauth2/appdev-token/<appID>
        url = '{0}/oauth2/appdev-token/{1}'.format(self.host, appID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data
    
    def post_oauth2_appdev_token(self, appID):
        """
        Generates a new token representing a user and a relayr Application.
        
        :param appID: The application's UUID.
        :type appID: string
        :rtype: A dict with fields describing the token.
        """
    
        # https://api.relayr.io/oauth2/appdev-token/<appID>
        url = '{0}/oauth2/appdev-token/{1}'.format(self.host, appID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def delete_oauth2_appdev_token(self, appID):
        """
        Revokes token.

        :param appID: The application's UUID.
        :type appID: string
        """
        
        # https://api.relayr.io/oauth2/appdev-token/<appID>
        url = '{0}/oauth2/appdev-token/{1}'.format(self.host, appID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    # ..............................................................................
    # Users
    # ..............................................................................

    def get_oauth2_user_info(self):
        """
        Returns information about the user initiating the request.
        
        :rtype: A dictionary with fields describing the user.
    
        Sample result (partly anonymized values)::
        
            {
                "email": "joe@foo.com", 
                "id": "...",
                "name": "joefoo"
            }
        """

        # https://api.relayr.io/oauth2/user-info
        url = '{0}/oauth2/user-info'.format(self.host)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def patch_user(self, userID, name=None, email=None):
        """
        Updates one or more user attributes.
        
        :param userID: the users's UUID
        :type userID: string
        :param name: the user name to be set
        :type name: string
        :param email: the user email to be set
        :type email: string
        :rtype: dict with user info fields
        """
    
        data = {}
        if name is not None:
            data.update(name=name)
        if email is not None:
            data.update(email=email)

        # https://api.relayr.io/users/%s
        url = '{0}/users/{1}'.format(self.host, userID)
        _, data = self.perform_request('PATCH', url, data=data, headers=self.headers)
        return data

    def post_user_app(self, userID, appID):
        """
        Installs a new app under a specific user.

        :param userID: the users's UUID
        :type userID: string
        :param appID: The application's UUID.
        :type appID: string
        """

        # https://api.relayr.io/users/%s/apps/%s
        url = '{0}/users/{1}/apps/{2}'.format(self.host, userID, appID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def delete_user_app(self, userID):
        """
        Uninstalls an app of a specific user.

        :param userID: the users's UUID
        :type userID: string
        """

        # https://api.relayr.io/users/%s/apps/%s
        url = '{0}/users/{1}/apps/{2}'.format(self.host, userID, appID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    def get_user_publishers(self, userID):
        """
        Returns all publishers owned by a specific user.
    
        :param userID: the users's UUID
        :type userID: string
        :rtype: list of dicts representing publishers
        """
        
        # https://api.relayr.io/users/%s/publishers
        url = '{0}/users/{1}/publishers'.format(self.host, userID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_user_apps(self, userID):
        """
        Returns all apps installed under a specific user.

        :param userID: the users's UUID
        :type userID: string
        :rtype: list of dicts ... with IDs and secrets
        """

        # https://api.relayr.io/users/%s/apps
        url = '{0}/users/{1}/apps'.format(self.host, userID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_user_transmitters(self, userID):
        """
        Returns all transmitters under a specific user.
    
        :param userID: the users's UUID
        :type userID: string
        :rtype: list of dicts with IDs and secrets
        """
    
        # https://api.relayr.io/users/%s/transmitters
        url = '{0}/users/{1}/transmitters'.format(self.host, userID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_user_devices(self, userID):
        """
        Returns all devices registered under a specific user.
    
        :param userID: the users's UUID
        :type userID: string
        :rtype: list of dicts ...
        """
                
        # https://api.relayr.io/users/%s/devices
        url = '{0}/users/{1}/devices'.format(self.host, userID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_user_devices_filtered(self, userID, meaning):
        """
        Returns all devices registered under a specific user filtered by meaning.
    
        :param userID: the users's UUID
        :type userID: string
        :param meaning: a meaning used for filtering results
        :type meaning: string
        :rtype: list of dicts representing devices
        """

        # https://api.relayr.io/users/%s/devices?meaning=%s
        url = '{0}/users/{1}/devices?meaning={2}'.format(self.host, userID, meaning)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_user_devices_bookmarks(self, userID):
        """
        Returns a list of devices bookmarked by the specific user.
    
        :param userID: the users's UUID
        :type userID: string
        :rtype: list of dicts, each representing a device

        Sample result (anonymized IDs)::

            [{u'firmwareVersion': u'1.0.0',
              u'id': '...',
              u'model': '...',
              u'name': 'My Wunderbar Microphone',
              u'owner': '...',
              u'public': True,
              u'secret': '238885'}]
        """
                
        # https://api.relayr.io/users/%s/devices/bookmarks
        url = '{0}/users/{1}/devices/bookmarks'.format(self.host, userID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def post_user_devices_bookmark(self, userID, deviceID):
        """
        Bookmarks a specific public device for a user.
    
        :param userID: the users's UUID
        :type userID: string
        :param deviceID: the UUID of the device to be bookmarked
        :type deviceID: string
        :rtype: list of dicts ...

        Sample result (anonymized IDs)::

            {'createdAt': '2014-11-05T16:31:06.429Z',
             'deviceId': '...',
             'userId': '...'}
         """
                
        # https://api.relayr.io/users/%s/devices/bookmarks
        url = '{0}/users/{1}/devices/{2}/bookmarks'.format(self.host, userID, deviceID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def delete_user_devices_bookmark(self, userID, deviceID):
        """
        Deletes a bookmark for a user and a device.

        :param userID: the users's UUID
        :type userID: string
        :param deviceID: the device's UUID
        :type deviceID: string
        :rtype: None
        """

        # https://api.relayr.io/users/%s/devices/%s/bookmarks
        url = '{0}/users/{1}/devices/{2}/bookmarks'.format(self.host, userID, deviceID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    def post_user_wunderbar(self, userID):
        """
        Returns the IDs and Secrets of the Master Module and Sensor Modules.
    
        :param userID: the users's UUID
        :type userID: string
        :rtype: dict with information about master and sensor modules/devices
    
        Sample result (abbreviated, some values anonymized)::
    
            {
                "bridge": { ... }, 
                "microphone": {
                    "name": "My Wunderbar Microphone", 
                    "public": False, 
                    "secret": "......", 
                    "owner": "...", 
                    "model": {
                        "readings": [
                            {
                                "meaning": "noise_level", 
                                "unit": "dba"
                            }
                        ], 
                        "manufacturer": "Relayr GmbH", 
                        "id": "...", 
                        "name": "Wunderbar Microphone"
                    }, 
                    "id": "...", 
                    "firmwareVersion": "1.0.0"
                }, 
                "light": { ... }, 
                "masterModule": {
                    "owner": "...", 
                    "secret": "............", 
                    "id": "...", 
                    "name": "My Wunderbar Master Module"
                }, 
                "infrared": { ... }, 
                "thermometer": { ... }, 
                "gyroscope": { ... }
            }    
        """
    
        # https://api.relayr.io/users/%s/wunderbar
        url = '{0}/users/{1}/wunderbar'.format(self.host, userID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def post_users_destroy(self, userID):
        """
        Removes all Wunderbars associated with a specific user.
    
        :param userID: the users's UUID
        :type userID: string
        """
    
        # https://api.relayr.io/users/%s/destroy-everything-i-love
        url = '{0}/users/{1}/destroy-everything-i-love'.format(self.host, userID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    # ..............................................................................
    # Applications
    # ..............................................................................

    def get_public_apps(self):
        """
        Returns a list of all applications on the relayr platform
    
        :rtype: list of dicts, each representing a relayr application
        """

        # https://api.relayr.io/apps
        url = '{0}/apps'.format(self.host)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def post_app(self, appName, publisherID, redirectURI, appDescription):
        """
        Registers a new app to the relayr platform.
    
        :rtype: list of dicts, each representing a relayr application
        """

        data = {
          "name": appName,
          "publisher": publisherID,
          "redirectUri": redirectURI,
          "description": appDescription
        }
        # https://api.relayr.io/apps
        url = '{0}/apps'.format(self.host)
        _, data = self.perform_request('POST', url, data=data, headers=self.headers)
        return data

    def get_app_info(self, appID):
        """
        Returns information about the app with the given ID.
        
        Sample result (anonymized token value)::
        
            {
                "id": "...",
                "name": "My App",
                "description": "My Wunderbar app",
                ...
            }
        """
    
        # https://api.relayr.io/apps/<appID>
        url = '{0}/apps/{1}'.format(self.host, appID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_app_info_extended(self, appID):
        """
        Returns extended information about the app with the given ID.
        
        Sample result (some values anonymized)::
        
            {
                "id": "...",
                "name": "My App",
                "publisher": "...",
                "clientId": "...",
                "clientSecret": "...",
                "description": "My Wunderbar app",
                "redirectUri": https://relayr.io
            }
        """
    
        # https://api.relayr.io/apps/<appID>/extended
        url = '{0}/apps/{1}/extended'.format(self.host, appID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data
    
    
    def patch_app(self, appID, description=None, name=None, redirectUri=None):
        """
        Updates one or more app attributes.
        
        :param appID: the application's UID
        :type appID: string
        :param description: the user name to be set
        :type description: string
        :param name: the user email to be set
        :type name: string
        :param redirectUri: the redirect URI to be set
        :type redirectUri: string
        
        Sample result (some values anonymized)::
        
            {
                "id": "...",
                "name": "My App",
                "publisher": "...",
                "clientId": "...",
                "clientSecret": "...",
                "description": "My Wunderbar app",
                "redirectUri": https://relayr.io
            }
        """

        data = {}
        if name is not None:
            data.update(name=name)
        if description is not None:
            data.update(description=description)
        if redirectUri is not None:
            data.update(redirectUri=redirectUri)

        # https://api.relayr.io/apps/<appID>
        url = '{0}/apps/{1}'.format(self.host, appID)
        _, data = self.perform_request('PATCH', url, data=data, headers=self.headers)
        return data

    def delete_app(self, appID):
        """
        Deletes an app from the relayr platform.
        
        :param appID: the application's UID
        :type appID: string
        
        """

        # https://api.relayr.io/apps/<appID>
        url = '{0}/apps/{1}'.format(self.host, appID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    def get_oauth2_app_info(self):
        """
        Returns info about the app initiating the request (the one in the token).

        Sample result (anonymized token value)::
        
            {
                "id": "...",
                "name": "My App",
                "description": "My Wunderbar app"
            }
        """

        # https://api.relayr.io/oauth2/app-info
        url = '{0}/oauth2/app-info'.format(self.host)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def post_app_device(self, appID, deviceID):
        """
        Connects an app to a device. Credentials are returned as part of the response.
        """
        # {{relayrAPI}}/apps/{{appID}}/devices/{{deviceID}}
        url = '{0}/apps/{1}/devices/{2}'.format(self.host, appID, deviceID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def delete_app_device(self, appID, deviceID):
        """
        Disconnect an app from a device.
        """
        # {{relayrAPI}}/apps/{{appID}}/devices/{{deviceID}}
        url = '{0}/apps/{1}/devices/{2}'.format(self.host, appID, deviceID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    # ..............................................................................
    # Publishers
    # ..............................................................................
    
    def get_public_publishers(self):
        """
        Returns a list of all publishers on the relayr platform.
    
        :rtype: list of dicts, each representing a relayr publisher
        """
    
        # https://api.relayr.io/publishers
        url = '{0}/publishers'.format(self.host)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data
        
    def post_publisher(self, userID, name):
        """
        Registers a new publisher.
    
        :param userID: the user UID of the publisher
        :type userID: string
        :param name: the publisher name
        :type name: string
        :rtype: a dict with fields describing the new publisher
        """
    
        # https://api.relayr.io/publishers
        data = {'owner': userID, 'name': name}
        url = '{0}/publishers'.format(self.host)
        _, data = self.perform_request('POST', url, data=data, headers=self.headers)
        return data

    def delete_publisher(self, publisherID):
        """
        Deletes a specific publisher from the relayr platform.
    
        :param publisherID: the publisher UID
        :type publisherID: string
        :rtype: an empty dict(?)
        """

        # https://api.relayr.io/publishers
        url = '{0}/publishers/{1}'.format(self.host, publisherID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    def get_publisher_apps(self, publisherID):
        """
        Returns a list of apps published by a specific publisher.
        
        :rtype: A list of apps.
        """

        # https://api.relayr.io/publishers/<id>/apps
        url = '{0}/publishers/{1}/apps'.format(self.host, publisherID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_publisher_apps_extended(self, publisherID):
        """
        Returns a list with extended information about the publisher's apps.
        
        :rtype: A list of apps.
        """

        # https://api.relayr.io/publishers/<id>/apps/extended
        url = '{0}/publishers/{1}/apps/extended'.format(self.host, publisherID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data
    
    
    def patch_publisher(self, publisherID, name=None):
        """
        Updates one or more publisher attributes.
        
        :param publisherID: the publisher's UID
        :type publisherID: string
        :param name: the publisher name to be set
        :type name: string
        :rtype: ``True``, if successful, ``False`` otherwise
        """
    
        data = {}
        if name is not None:
            data.update(name=name)
        
        # https://api.relayr.io/publishers/<id>
        url = '{0}/publishers/{1}'.format(self.host, publisherID)
        _, data = self.perform_request('PATCH', url, data=data, headers=self.headers)
        return data

    # ..............................................................................
    # Devices
    # ..............................................................................
    
    def get_device_configuration(self, deviceID):
        """
        Returns the device's configuration, default values and schema.

        Example result::

            {
                "version": "1.0.0", 
                "configuration": {
                    "defaultValues": {
                        "frequency": 1000
                    }, 
                    "schema": {
                        "required": [
                            "frequency"
                        ], 
                        "type": "object", 
                        "properties": {
                            "frequency": {
                                "minimum": 5, 
                                "type": "integer", 
                                "description": "Frequency of the sensor updates in milliseconds"
                            }
                        }, 
                        "title": "Relayr configuration schema"
                    }
                }
            }

        :param deviceID: the device UID
        :type deviceID: string
        """
        
        # https://api.relayr.io/devices/<deviceID>/firmware
        url = '{0}/devices/{1}/firmware'.format(self.host, deviceID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def post_device_configuration(self, deviceID, frequency):
        """
        Modifies the configuration of a specific device facillitated by a schema.

        :param deviceID: the device UID
        :type deviceID: string
        :param frequency: the number of ms between two sensor transmissions
        :type frequency: integer
        """
        
        data = {'frequency': frequency}
        # https://api.relayr.io/devices/<deviceID>/configuration
        url = '{0}/devices/{1}/configuration'.format(self.host, deviceID)
        _, data = self.perform_request('POST', url, data=data, headers=self.headers)
        return data

    def get_public_devices(self, meaning=''):
        """
        Returns list of all public devices on the relayr platform.
    
        :param meaning: required meaning in the device model's ``readings`` attribute
        :type meaning: string
        :rtype: list of dicts, each representing a relayr device
        """
        
        # https://api.relayr.io/devices/public
        url = '{0}/devices/public'.format(self.host)
        if meaning:
            url += '?meaning={0}'.format(meaning)
        _, data = self.perform_request('GET', url)
        return data

    def post_device(self, name, owner, modelId, firmwareVersion):
        """
        Registers a new device on the relayr platform.
    
        :rtype: list of dicts, each representing a relayr device
        """

        data = {
          "name": name,
          "owner": owner,
          "model": modelId,
          "firmwareVersion": firmwareVersion
        }
        # https://api.relayr.io/devices
        url = '{0}/devices'.format(self.host)
        _, data = self.perform_request('POST', url, data=data, headers=self.headers)
        return data

    def get_device(self, deviceID):
        """
        Returns information about a specific device.
            
        :param deviceID: the device UID
        :type deviceID: string
        :rtype: a dict with fields containing information about the device
    
        Raises ``exceptions.RelayrApiException`` for invalid UIDs or missing 
        credentials...
        """
        
        # https://api.relayr.io/devices/%s
        url = '{0}/devices/{1}'.format(self.host, deviceID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def patch_device(self, deviceID=None, deviceName=None, deviceDescription=None, deviceModel=None, isDevicePublic=None):
        """
        Updates one or more device attributes.
            
        :param deviceID: the device UID
        :type deviceID: string
        :rtype: a dict with fields containing information about the device
    
        Raises ``exceptions.RelayrApiException`` for invalid UIDs or missing 
        credentials...
        """

        data = {
          "name": deviceName,
          "description": deviceDescription,
          "model": deviceModel,
          "public": isDevicePublic
        }
        for k, v in data.items():
            if v is None:
                del data[k]
        # https://api.relayr.io/devices/%s
        url = '{0}/devices/{1}'.format(self.host, deviceID)
        _, data = self.perform_request('PATCH', url, data=data, headers=self.headers)
        return data

    def delete_device(self, deviceID):
        """
        Deletes a device from the relayr plarform.

        :param deviceID: the device UID
        :type deviceID: string
        :rtype: a dict with fields containing information about the device
        """
        # https://api.relayr.io/devices/%s
        url = '{0}/devices/{1}'.format(self.host, deviceID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    def get_device_apps(self, deviceID):
        """
        Returns all the apps connected to a specific device.
    
        :param deviceID: the device UID
        :type deviceID: string
        :rtype: a list of dicts with information about apps
        """
    
        # https://api.relayr.io/devices/<deviceID>/apps
        url = '{0}/devices/{1}/apps'.format(self.host, deviceID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    ## TODO: remove in version 0.3.0
    def post_devices_supscription(self, deviceID):
        if config.LOG:
            self.logger.info("Deprecated method 'post_devices_supscription' called. Please use 'post_devices_public_subscription' instead.")
        self.post_devices_public_subscription(deviceID)

    def post_devices_public_subscription(self, deviceID):
        """
        Subscribes a user to a public device.
    
        :param deviceID: the device's UID
        :type deviceID: string
        :rtype: dict with connection credentials
        
        Sample result (anonymized values)::
    
            {
                "authKey": "...", 
                "cipherKey": "...", 
                "channel": "...", 
                "subscribeKey": "sub-c-..."
            }
        """
    
        # https://api.relayr.io/devices/%s/subscription
        url = '{0}/devices/{1}/subscription'.format(self.host, deviceID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def post_devices_subscription(self, appID, deviceID):
        """
        Subscribes a user to a device.
        Sample result (anonymized values)::

            {
                "authKey": "...",
                "cipherKey": "...",
                "channel": "...",
                "subscribeKey": "sub-c-..."
            }
        """

        # https://api.relayr.io/apps/%s/devices/%s/subscription
        url = '{0}/apps/{1}/devices/{2}/subscription'.format(self.host, appID, deviceID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def post_device_command(self, deviceID, command, data):
        """
        Sends a command to a specific device.
        """
        # https://api.relayr.io/devices/<deviceID>/cmd/<command>
        url = '{0}/devices/{1}/cmd/{2}'.format(self.host, deviceID, command)
        _, data = self.perform_request('POST', url, data=data, headers=self.headers)
        return data

    def post_device_app(deviceID, appID):
        """
        Connects a device to an app. 
        
        Credentials for data reception are returned as part of the response.
        """
        # {{relayrAPI}}/devices/{{deviceID}}/apps/{{appID}}
        url = '{0}/devices/{1}/apps/{2}'.format(self.host, deviceID, appID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def delete_device_app(deviceID, appID):
        """
        Disconnects a device from an app.
        """
        # {{relayrAPI}}/devices/{{deviceID}}/apps/{{appID}}
        url = '{0}/devices/{1}/apps/{2}'.format(self.host, appID, deviceID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    # ..............................................................................
    # Device models
    # ..............................................................................
    
    def get_public_device_models(self):
        """
        Returns list of all device models available on the relayr platform.
    
        :rtype: list of dicts, each representing a relayr device model
        """
    
        # https://api.relayr.io/device-models
        url = '{0}/device-models'.format(self.host)
        _, data = self.perform_request('GET', url)
        return data

    def get_device_model(self, devicemodelID):
        """
        Returns information about a specific device model.
        
        :rtype: A nested dictionary structure with fields describing the DM.
        """
    
        # https://api.relayr.io/device-models/<id>
        url = '{0}/device-models/{1}'.format(self.host, devicemodelID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def get_public_device_model_meanings(self):
        """
        Returns list of all device model meanings (no credentials needed).
    
        :rtype: list of dicts, each representing a relayr device model meaning
        """

        # https://api.relayr.io/device-models/meanings
        url = '{0}/device-models/meanings'.format(self.host)
        _, data = self.perform_request('GET', url)
        return data

    # ..............................................................................
    # Transmitters
    # ..............................................................................

    def get_transmitter(self, transmitterID):
        """
        Returns information about a specific transmitter with the given ID.
    
        :param transmitterID: the transmitter UID
        :type transmitterID: string
        :rtype: a dict with fields describing the transmitter
        """
    
        # https://api.relayr.io/transmitters/<id>
        url = '{0}/transmitters/{1}'.format(self.host, transmitterID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def post_transmitter(self, transmitterID, owner=None, name=None):
        """
        Registers a new Transmitter on the relayr platform.
    
        :param transmitterID: the transmitter UID
        :type transmitterID: string
        :param owner: the transmitters owner's UID
        :type owner: string
        :param name: the transmitter name
        :type name: string
        :rtype: an empty dict(?)
        """

        data = {}
        if owner is not None:
            data.update(owner=owner)
        if name is not None:
            data.update(name=name)
        
        # https://api.relayr.io/transmitters/<id>
        url = '{0}/transmitters/{1}'.format(self.host, transmitterID)
        _, data = self.perform_request('POST', url, data=data, headers=self.headers)
        return data

    def patch_transmitter(self, transmitterID, name=None):
        """
        Updates information about a specific transmitter.
    
        :param transmitterID: the transmitter UID
        :type transmitterID: string
        :param name: the transmitter name
        :type name: string
        :rtype: an empty dict(?)
        """

        data = {}
        if name is not None:
            data.update(name=name)
                
        # https://api.relayr.io/transmitters/<id>
        url = '{0}/transmitters/{1}'.format(self.host, transmitterID)
        _, data = self.perform_request('PATCH', url, data=data, headers=self.headers)
        return data

    def delete_transmitter(self, transmitterID):
        """
        Deletes a specific transmitter from the relayr platform.
    
        :param transmitterID: the transmitter UID
        :type transmitterID: string
        :rtype: an empty dict(?)
        """

        # https://api.relayr.io/transmitters/<id>
        url = '{0}/transmitters/{1}'.format(self.host, transmitterID)
        _, data = self.perform_request('DELETE', url, headers=self.headers)
        return data

    def post_transmitter_device(self, transmitterID, deviceID):
        """
        Connects a transmitter to a device.
    
        :param transmitterID: the transmitter UID
        :type transmitterID: string
        :param deviceID: the device UID
        :type deviceID: string
        :rtype: an empty dict(?)
        """
    
        # https://api.relayr.io/transmitters/<transmitterID>/devices/<deviceID>
        url = '{0}/transmitters/{1}/devices/{2}'.format(self.host, transmitterID, deviceID)
        _, data = self.perform_request('POST', url, headers=self.headers)
        return data

    def get_transmitter_devices(self, transmitterID):
        """
        Returns a list of devices connected to a specific transmitter.
    
        :param transmitterID: the transmitter UID
        :type transmitterID: string
        :rtype: a list of devices
        """
    
        # https://api.relayr.io/transmitters/<transmitterID>/devices
        url = '{0}/transmitters/{1}/devices'.format(self.host, transmitterID)
        _, data = self.perform_request('GET', url, headers=self.headers)
        return data

    def delete_transmitter_device(self, transmitterID, deviceID):
        """
        Disconnects a transmitter from a device.
    
        :param transmitterID: the transmitter UID
        :type transmitterID: string
        :param deviceID: the device UID
        :type deviceID: string
        :rtype: an empty dict(?)
        """
    
        # https://api.relayr.io/transmitters/<transmitterID>/devices/<deviceID>
        url = '{0}/transmitters/{1}/devices/{2}'.format(self.host, transmitterID, deviceID)
        _, data = self.perform_request('DELETE', url, data=data, headers=self.headers)
        return data



"This is for testing HTTP-POSTing data using requests with either dicts or lists."

import json

import pytest

from relayr.resources import Device


class TestHttpPostData(object):
    "Test HTTP post requests with different types of data payloads."
    # http://curl.haxx.se/docs/httpscripting.html

    def test_perform_request(self, fix_registered):
        "Test building cURL calls."
        token = fix_registered.testset1['token']
        from relayr import Client
        c = Client(token=token)
        headers = {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer 3A1owTXfLKRew0s.BF-4qhEdXQTHgDr4',
            'User-Agent': 'io.relayr.sdk.python/0.2.1 (Darwin-10.8.0; i386-64bit; CPython-2.7.6)'
        }
        res = c.api.perform_request('get', 'https://api.relayr.io/devices/a3386628-61a8-4a80-9a42-1bd3c64c46a5/cmd/led', 
            data={'cmd': 1},
            headers=headers)
        # import pdb; pdb.set_trace();

    def test_build_curl_0(self, fix_registered):
        "Test building cURL calls."
        token = fix_registered.testset1['token']
        from relayr.api import build_curl_call
        res = build_curl_call('post', 'http://foo.com')
        assert res == 'curl -X POST http://foo.com'
        res = build_curl_call('post', 'http://foo.com', headers={'foo': 'bar', 'bar': 'foo'})
        assert res == 'curl -X POST http://foo.com -H "foo: bar" -H "bar: foo"'
        res = build_curl_call('post', 'http://foo.com', data={'foo': 'bar', 'bar': 'foo'})
        assert res == 'curl -X POST http://foo.com --data foo=bar&bar=foo'
        res = build_curl_call('post', 'http://foo.com', data=json.dumps([{'foo': 'bar'}]))
        assert res == 'curl -X POST http://foo.com --data \'[{"foo": "bar"}]\''

    def _test_post_dict(self, fix_registered):
        "Test building a cURL call."
        token = fix_registered.testset1['token']
        from relayr import Client
        c = Client(token=token)
        dev.send_command('led', data)

    def _test_post_list(self, fix_registered):
        "Post list as JSON."
        token = fix_registered.testset1['token']
        from datetime import datetime
        from relayr import Client
        c = Client(token=token)
        messages = [
            {
                "timestamp": datetime.now().isoformat(),
                "message": "testing xyz",
                "connection": { "internet": True}
            },
            {
                "timestamp": datetime.now().isoformat(),
                "message": "testing xyz",
                "connection": { "internet": True}
            },
        ]
        res = c.api.post_client_log(messages)



# -*- coding: utf-8 -*-

"""
Functionality for working with Bluetooth, especially LE (Low Energy).

Since PyBluez doesn't support Bluetooth LE yet this module provides an
abstraction on the tools ``hcitool`` and ``gatttool``.

This module needs to be run with root permisions.

Example:

.. code-block:: python

    >>> from ble import scan_ble_devices
    >>> scan_ble_devices()
    [{'addr': 'F4:D7:AB:6D:EB:D5', 'name': 'WunderbarMIC'},
     {'addr': 'C3:18:9B:AF:F6:F1', 'name': 'WunderbarBRIDG'},
     ...
     {'addr': 'F2:EE:50:45:39:74', 'name': 'WunderbarGYRO'}]
"""

import re
import sys
import time
import shlex
import subprocess

import pexpect


# Device information service
Device_Name_UUID = '2a00'
Manufacturer_Name_UUID = '2a29'
Firmware_Revision_UUID = '2a26'
Hardware_Revision_UUID = '2a27'
# Battery service
Battery_Level_UUID = '2a19'


def data2str(data):
    """
    Convert some data to a string.

    An empty or None value is returned unchanged (helpful for testing), e.g.:

    '57 75 6e 64 65 72 62 61 72 49 52' -> 'WunderbarIR'
    '' -> ''
    """
    if not data: return data
    text = ''.join([chr(int(v, 16)) for v in data.split()])
    return text


def str2data(str):
    """
    Convert a string to some data bytes.

    An empty or None value is returned unchanged (helpful for testing), e.g.:

    'WunderbarIR' -> '57 75 6e 64 65 72 62 61 72 49 52'
    '' -> ''
    """
    if not str: return str
    data = ' '.join([hex(ord(c))[2:] for c in str])
    return data


def scan_ble_devices(hci_name='hci0', name_filter='.*', timeout=1):
    """
    Return list of discovered Bluetooth LE devices with address and name.
    
    Attention: This must be run with sudo permissions!

    :param hci_name: the HCI device name
    :type hci_name: string
    :param name_filter: a regular expression that service names need to match
    :type name_filter: string
    :param timeout: the timeout value for the scan duration in seconds
    :type timeout: string
    :rtype: a list of dicts, one per device, each with a name and addr entry

    Sample output::
    
        [{'addr': 'F4:D7:AB:6D:EB:D5', 'name': 'WunderbarMIC'},
         {'addr': 'C3:18:9B:AF:F6:F1', 'name': 'WunderbarBRIDG'},
         {'addr': 'D0:DA:36:13:EB:39', 'name': 'WunderbarIR'},
         {'addr': 'DE:7F:20:07:AD:04', 'name': 'WunderbarHTU'},
         {'addr': 'F2:84:DC:44:D7:5C', 'name': 'WunderbarLIGHT'},
         {'addr': 'E9:5D:10:B9:E2:1C', 'name': 'WunderbarMIC'},
         {'addr': 'F2:EE:50:45:39:74', 'name': 'WunderbarGYRO'}]
    """
    
    conn = pexpect.spawn('hciconfig %s reset' % hci_name) # needs sudo
    time.sleep(0.2)

    conn = pexpect.spawn('timeout %d hcitool lescan' % timeout) # needs sudo
    time.sleep(0.2)

    conn.expect('LE Scan \.+', timeout=timeout)

    output = ''
    line_pat = '(?P<addr>([0-9A-F]{2}:){5}[0-9A-F]{2}) (?P<name>.*)'
    
    while True:
        try:
            res = conn.expect(line_pat)
            output += conn.after
        except pexpect.EOF:
            break
      
    lines = re.split('\r?\n', output.strip())
    lines = list(set(lines))
    lines = [line for line in lines if re.match(line_pat, line)]
    lines = [re.match(line_pat, line).groupdict() for line in lines]
    lines = [line for line in lines if re.match(name_filter, line['name'])]
    
    return lines


class GattDevice(object):
    """
    A class to communicate via GATT with a Bluetooth LE device.

    This class uses the Bluez ``getttool`` in a non-interactive manner.
    """
    
    def __init__(self, addr):
        """
        Instantiate an object representing a device accessible via GATT.

        :param addr: the MAC address of the device
        :type addr: string
        """

        self.addr = addr
        self.data = {
            'services': [],
            'characteristics': [],
            'char-desc': [],
            'read_log': [], # log of read data: {'ts':..., ''handle':..., 'bytes':...}
            'write_log': []  # log of written data: same...
        }
        self.callbacks = {}

    def primary(self):
        """
        Discover primary services.
        """

        # example line format:
        # 'attr handle = 0x0001, end grp handle = 0x0007 uuid: 00001800-0000-1000-8000-00805f9b34fb'
        cmd = 'gatttool -t random -b %s --primary' % self.addr
        res = subprocess.check_output(shlex.split(cmd))
        lines = re.split('\r?\n', res)
        lines = sorted(set(lines))
        self.data['services'] = lines
        return lines

    def characteristics(self, uuid=None):
        """
        Read list of characteristics.

        If the ``uuid`` parameter is given, only this characteristic will be read.

        :param uuid: uuid to be read
        :type uuid: string
        """

        # example line format:
        # 'handle = 0x0002, char properties = 0x02, char value handle = 0x0003, uuid = 00002a00-0000-1000-8000-00805f9b34fb'
        cmd = 'gatttool -t random -b %s --characteristics' % self.addr
        if uuid:
            cmd += ' --uuid=%s' % uuid
        res = subprocess.check_output(shlex.split(cmd))
        lines = re.split('\r?\n', res)
        pat = '^handle\s*=\s*(?P<name_handle>0x[0-9a-fA-F]{4}),\s*char\s*properties\s*=\s*(?P<properties>0x[0-9a-fA-F]+),\s*char\s*value\s*handle\s*=\s*(?P<value_handle>0x[0-9a-fA-F]{4}),\s*uuid\s*=\s*(?P<uuid>[0-9a-fA-F\-]+)'
        lines = [re.match(pat, line).groupdict() for line in lines]
        for line in lines:
            short = line['uuid'].split('-')[0]
            while short[0] == '0': short = short[1:]
            line['_uuid'] = short
        self.data['characteristics'] = lines
        return lines

    def char_desc(self):
        """
        Read list of characteristic descriptors.
        """

        # example line format:
        # 'handle = 0x000e, uuid = 00002010-0000-1000-8000-00805f9b34fb'
        cmd = 'gatttool -t random -b %s --char-desc' % self.addr
        res = subprocess.check_output(shlex.split(cmd))
        lines = re.split('\r?\n', res)
        # lines = re.findall('handle: .* uuid: [0-9a-f\-]{36}', res)
        self.data['char-desc'] = lines
        return lines

    def char_read_hnd(self, handle):
        """
        Read a data string representing the value of a handle.

        Return None if the handle cannot be read, e.g. because it is invalid.

        :param handle: uuid to be read, e.g. '0x0017'
        :type handle: string
        """

        cmd = 'gatttool -t random -b %s --char-read --handle %s' % (self.addr, handle)
        res = subprocess.check_output(shlex.split(cmd))
        if not res.startswith('Characteristic value/descriptor:'):
            return None
        res = re.match('Characteristic value/descriptor: (.*)', res).groups()[0]
        byte_values = res # .split()
        return byte_values

    # higher-level interface

    def read_device_name(self):
        "Return device name."

        for char in self.data['characteristics']:
            if char['_uuid'] == Device_Name_UUID:
                value_handle = char['value_handle']
        data = self.char_read_hnd(value_handle) # eg. '57 75 6e 64 65 72 62 61 72 49 52'
        name = ''.join([chr(int(v, 16)) for v in data.split()]) # e.g. 'WunderbarIR'
        return name

    def read_battery_level(self):
        "Return current battery level."

        value_handle = None
        for char in self.data['characteristics']:
            if char['_uuid'] == Battery_Level_UUID:
                value_handle = char['value_handle']
        if not value_handle:
            return None
        data = self.char_read_hnd(value_handle) # eg. '64'
        level = int(data, 16) # e.g. 100
        return level


class WunderbarGattDevice(GattDevice):
    """
    A class to communicate via GATT with a Bluetooth LE Wunderbar device.

    This class uses the Bluez ``getttool`` in a non-interactive manner.
    """

    def switch_led_on(self):
        "Switch device LED on."

        pass

    def switch_led_off(self):
        "Switch device LED off."

        pass

    def read_value_named(self, name):
        "Return value with given name."

        pass

    def write_value_named(self, name, value):
        "Write value with given name."

        pass


class DeviceCallbacks(object):
    "A set of callbacks for handling device data." 

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Tests for accessing Relayr Wunderbar devices via Bluetooth LE.

Some of these test need root permissions, see docstrings below.
"""

import sys
import subprocess

import pytest


ON_LINUX = sys.platform.startswith("linux")
HAVE_BLUEZ = subprocess.Popen(['which', 'hcitool']).stdout != ''


@pytest.mark.skipif(not ON_LINUX, reason="requires Linux")
class TestBle(object):
    "Tests for accessing BLE devices via Bluez tools (needs root permissions)."

    def test_ble_scan(self):
        "Test BLE service discovery (needs switched on BLE devices!)."
        from relayr.ble import scan_ble_devices
        devs = scan_ble_devices(timeout=3)
        assert len(devs) > 0

    def test_access_device(self, fix_registered):
        "Test accessing BLE device."
        from relayr.ble import WunderbarGattDevice
        addr = fix_registered.testset4['bluetoothAddr']
        d = WunderbarGattDevice(addr)
        assert len(d.primary()) > 0
        assert len(d.characteristics()) > 0
        assert len(d.char_desc()) > 0
        assert len(d.char_read_hnd(0x0016)) > 0



"""
A set of variables used for testing the relayr API.
Attention: You must provide your own eaningful values 
before you can run large parts of the test suite!

Get 'token' on relayr.io!

Remember that tokens expire after some time, check the documentation!
"""

testset1 = {
    'userID': '...', 
    'userName': '...', 
    'userEmail': '...',
    'relayrAPI': 'https://api.relayr.io',
    'redirectURI': 'https://relayr.io',
    'response_type': 'code',
    'clientId': '...',
    'clientSecret': '...',
    'scope': 'access-own-user-info',
    'appID': '...',
    'appName': '...',
    'publisherID': '...',
    'deviceID': '...',
    'transmitterID': '...',
    'transmitterName': '...',
    'description': "...",
    'useragent': '...',
    'token': '...',
}

testset2 = {
    'publisherID': '...',
}

testset3 = {
    'PUBLISH_KEY': '...',
    'SUBSCRIBE_KEY': '...',
    'channel': '...',
    'message': '...'
}

testset4 = {
    'bluetoothAddr': '...',
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

from setuptools import setup, Command


PY_VERSION = sys.version_info[0]


class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys, subprocess
        errno = subprocess.call([sys.executable, 'runtests.py', 'tests'])
        raise SystemExit(errno)


exec(open('relayr/version.py').read())

with open('README.rst') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    install_requires = f.read().strip().split('\n')

if PY_VERSION == 2:
    with open('requirements_py2.txt') as f:
        install_requires += f.read().strip().split('\n')

tests_require = [
    # 'requests>=1.0.0, <3.0.0',
]


setup(
    name = "relayr",
    description = "Python client for Relayr API",
    license = "MIT",
    url = "https://github.com/relayr/python-sdk",
    long_description = long_description,
    version = __version__,
    author = "Relayr Team",
    author_email = "team@relayr.io",
    packages = ['relayr', 'relayr.utils'],
    keywords = ['relayr', 'rest', 'api', 'cloud', 'python', 'client', 'iot',
        'wunderbar'],
    classifiers = [
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Home Automation",
        "Topic :: Internet",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Hardware",
        "Topic :: System :: Networking",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
    ],
    install_requires = install_requires,
    tests_require = tests_require,
    cmdclass = {'test': PyTest},
    zip_safe = False
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Workaround functions and stuff...
"""

import json


def cleanup_pubnub_message_py3(message):
    """
    Return a cleaned version of a PubNub message under Python 3.

    Workaround for Pubnub in Python 3 passing messages appended with
    strange characters ('\x06', '\x07') plus outer quotes... making json
    barf.

    PY3: '"{\\"ts\\":1414672791632,\\"snd_level\\":25}"\x07\x07\x07\x07\x07\x07\x07'
    PY2: '\'{"ts":1414672791632,"snd_level":25}\''
    """
    low_ascii_chars = ''.join([chr(i) for i in range(8)])
    res = message.rstrip(low_ascii_chars)
    res = json.loads(res)

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Example script using the Wunderbar over the Relayr cloud.

This will connect to a microphone, read its noise level and send
an email notification to some receiver if that noise level exceeds
a certain threshold.
"""

import sys
import json
import time
import getpass
import smtplib
from email.mime.text import MIMEText

from relayr import Client


# Replace with your own values!
ACCESS_TOKEN = '...'
MICROPHONE_ID = '...'
# EMAIL/SMTP settings, please provide your own!
RECEIVER = '...'
SMTP_SERVER = '...'
SMTP_USERNAME = '...'
SMTP_PASSWORD = '' # will be requested at run time if left empty 

try:
    settings = [ACCESS_TOKEN, MICROPHONE_ID, RECEIVER, SMTP_SERVER, SMTP_USERNAME]
    assert not any(map(lambda x: x=='...', settings))
except AssertionError:
    print('Please provide meaningful settings in the code first!')
    sys.exit(1)


class Callbacks(object):
    "A class providing callbacks for incoming data from some device."

    def __init__(self, device):
        "An initializer to capture the device for later use."

        self.device = device

    def send_email(self, text):
        "Send an email notification."
        
        sender = 'Wunderbar <notification-noreply@relayr.io>'
        subject = 'Wunderbar Notification from Device: %s' % self.device.name
        msg = MIMEText(text)
        msg['Subject'] = subject
        msg['From'] = sender
        msg['To'] = RECEIVER
        s = smtplib.SMTP(SMTP_SERVER)
        prompt = "SMTP user password for user '%s'? " % SMTP_USERNAME
        global SMTP_PASSWORD
        SMTP_PASSWORD = SMTP_PASSWORD or getpass.getpass(prompt)
        s.login(SMTP_USERNAME, SMTP_PASSWORD)
        s.sendmail(sender, [RECEIVER], msg.as_string())
        s.quit()
        print("Email notification sent to '%s'" % RECEIVER)

    def microphone(self, message, channel):
        "Callback displaying incoming noise level data and email if desired."

        # Needs workaround handling a Pubnub strangeness on Python 3...
        # not added here for brevity (see relayr/utils/workarounds.py).
        level = json.loads(message)['snd_level'] 

        print(level)
        threshold = 75
        if level > threshold:
            dname, did = self.device.name, self.device.id
            text = "Notification from '%s' (%s):\n" % (dname, did)
            text += ("The noise level now is %d (> %d)! " % (level, threshold))
            text += "Put on a sound protection helmet before you get deaf!"
            self.send_email(text)


def connect():
    "Connect to a device and read data for some time."

    c = Client(token=ACCESS_TOKEN)
    m = c.get_device(id=MICROPHONE_ID).get_info()
    cb = Callbacks(m)
    user = c.get_user()
    app = c.get_app()
    conn = user.connect_device(app, m, cb.microphone)
    conn.start()
    print("Monitoring '%s' (%s) for 60 seconds..." % (m.name, m.id))
    try:
        time.sleep(60)
    except KeyboardInterrupt:
        print('')
    conn.stop()
    print("Stopped")


if __name__ == "__main__":
    connect()
